/* 

   ircleaner.c

   a study done with sane (Scanner Access Now Easy) in mind

   Copyright (C) 2011 Michael Rickmann <mrickma@gwdg.de>
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.

   histogram routines, threshold_yen and threshold_otsu were adapted
   from M. Emre Celebi's FOURIER 0.8 project, see
   http://sourceforge.net/projects/fourier-ipal

*/

#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <limits.h>
#include <values.h>
#include <math.h>
#include <pthread.h>
#include <sys/sysinfo.h>
#include <sys/time.h>
#include <pam.h>

#include "ircleaner.h"

#define RED_NAME "color-0.pnm"
#define GREEN_NAME "color-1.pnm"
#define BLUE_NAME "color-2.pnm"
#define IRED_NAME "color-3.pnm"
#define RGB1_NAME "0color-rgb.pnm"
#define RGB3_NAME "3dilate-rgb.pnm"
#define RGB4_NAME "4gamma-rgb.pnm"
#define CLEAN_NAME "1clean-3.pnm"
#define THRESH_1_NAME "2yen-threshold-3.pnm"
#define THRESH_2_NAME "2otsu-threshold-3.pnm"
#define MADM_NAME "2madmean-threshold-3.pnm"


Histogram *
alloc_histo (const HistoType type, const int num_bins)
{
  Histogram *histo = NULL;

  if (!type)
    {
      fprintf (stderr, "alloc_histo: Invalid histogram type !");
    }

  if (num_bins <= 0)
    {
      fprintf (stderr, "alloc_histo: Number of bins ( %d ) must be positive !",
	     num_bins);
      return NULL;
    }

  histo = malloc (sizeof (Histogram));
  histo->type = type;
  histo->num_bins = num_bins;
  histo->num_pixels = INT_MIN;	/* Illegal value */

  /* Allocate storage for histogram data */
  if (type == HISTO_INT)
    {
      histo->data.int_data = (int *) calloc (num_bins, sizeof (int));
      if (histo->data.int_data == NULL)
	{
	  ERROR_RET ("alloc_histo: Insufficient memory !", NULL);
	}
    }
  else
    {
      histo->data.double_data = (double *) calloc (num_bins, sizeof (double));
      if (histo->data.double_data == NULL)
	{
	  ERROR_RET ("alloc_histo: Insufficient memory !", NULL);
	}
    }

  return histo;
}

void *
get_histo_data (const Histogram * histo)
{
  if (histo->type == HISTO_INT)
    {
      return histo->data.int_data;
    }
  return histo->data.double_data;
}

Histogram *
create_histo (const Image * img)
{
  uint8_t *img_data8;
  uint16_t *img_data16;
  int i;
  int num_pixels;
  int *histo_data;
  Histogram *histo = NULL;

  if ((img->type != PIX_GRAY8) && (img->type != PIX_GRAY16))
    ERROR_RET ("create_histo: Not a grayscale image !", 0);

  /* Allocate storage for the histogram */
  histo = alloc_histo (HISTO_INT, NUM_GRAY);
  if (histo == NULL)
    ERROR_RET ("create_histo: alloc_histo() failed !", NULL);

  num_pixels = img->num_rows * img->num_cols;
  histo->num_pixels = num_pixels;
  histo_data = get_histo_data (histo);

  /* Populate the histogram */
  if (img->type == PIX_GRAY8)
    {
      img_data8 = img->data.sample8;
      for (i = num_pixels; i > 0; i--)
	histo_data[*img_data8++]++;
    }
  else
    {
      img_data16 = img->data.sample16;
      for (i = num_pixels; i > 0; i--)
	histo_data[*img_data16++ >> 8]++;
    }

  return histo;
}


Histogram *
normalize_histo (const Histogram * in_histo)
{
  int ih;
  int num_pixels;
  int num_bins;			/* number of bins in the histogram */
  int *in_data;			/* input histogram data */
  double term;
  double *out_data;		/* output histogram data */
  Histogram *out_histo;

  if (in_histo->type != HISTO_INT)
    {
      ERROR_RET ("normalize_histo: Not an integer histogram !", NULL);
    }

  num_pixels = in_histo->num_pixels;
  num_bins = in_histo->num_bins;
  in_data = get_histo_data (in_histo);

  /* Allocate storage for the histogram */
  out_histo = alloc_histo (HISTO_DBL, num_bins);
  if (out_histo == NULL)
    {
      ERROR_RET ("normalize_histo: alloc_histo() failed !", NULL);
    }

  out_histo->num_pixels = num_pixels;
  out_data = get_histo_data (out_histo);

  term = 1.0 / (double) num_pixels;

  /* Calculate the normalized histogram */
  for (ih = 0; ih < num_bins; ih++)
    {
      out_data[ih] = term * in_data[ih];
    }

  return out_histo;
}


void *
accumulate_histo (const Histogram * histo)
{
  int ih;
  int num_bins;

  num_bins = histo->num_bins;

  /* Codes for the INT and DBL histograms are nearly identical */

  if (histo->type == HISTO_INT)
    {
      int *histo_data;
      int *accum_data;

      histo_data = get_histo_data (histo);

      accum_data = (int *) malloc (num_bins * sizeof (int));
      if (accum_data == NULL)
	{
	  ERROR_RET ("accumulate_histo: Insufficient memory !", NULL);
	}

      accum_data[0] = histo_data[0];
      for (ih = 1; ih < num_bins; ih++)
	{
	  accum_data[ih] = accum_data[ih - 1] + histo_data[ih];
	}

      return accum_data;
    }
  else
    {
      double *histo_data;
      double *accum_data;

      histo_data = get_histo_data (histo);

      accum_data = (double *) malloc (num_bins * sizeof (double));
      if (accum_data == NULL)
	{
	  ERROR_RET ("accumulate_histo: Insufficient memory !", NULL);
	}

      accum_data[0] = histo_data[0];
      for (ih = 1; ih < num_bins; ih++)
	{
	  accum_data[ih] = accum_data[ih - 1] + histo_data[ih];
	}

      return accum_data;
    }

  /*@notreached@ */
}

void
free_histo (Histogram * histo)
{
  if (histo->type == HISTO_INT)
    {
      free (histo->data.int_data);
      histo->data.int_data = NULL;
    }
  else
    {
      free (histo->data.double_data);
      histo->data.double_data = NULL;
    }

  histo->type = HISTO_INVALID;
  histo->num_bins = INT_MIN;
  histo->num_pixels = INT_MIN;
}

int
threshold_yen (const Image * img)
{
  int ih, it;
  int threshold;
  double crit;
  double max_crit;
  double *data;			/* normalized histogram data */
  double *P1;			/* cumulative normalized histogram */
  double *P1_sq;		/* cumulative normalized histogram */
  double *P2_sq;		/* see below */
  Histogram *norm_histo;	/* normalized histogram */

  if ((img->type != PIX_GRAY8) && (img->type != PIX_GRAY16))
    {
      ERROR_RET ("Not a grayscale image !", INT_MIN);
    }

  /* Calculate the normalized histogram */
  norm_histo = normalize_histo (create_histo (img));
  if (norm_histo == 0)
    {
      ERROR_RET ("normalize_histo() failed !", INT_MIN);
    }

  data = get_histo_data (norm_histo);

  /* Calculate the cumulative normalized histogram */
  P1 = accumulate_histo (norm_histo);

  P1_sq = (double *) malloc (NUM_GRAY * sizeof (double));

  P1_sq[0] = data[0] * data[0];
  for (ih = 0 + 1; ih < NUM_GRAY; ih++)
    {
      P1_sq[ih] = P1_sq[ih - 1] + data[ih] * data[ih];
    }

  P2_sq = (double *) malloc (NUM_GRAY * sizeof (double));

  P2_sq[MAX_GRAY] = 0.0;
  for (ih = MAX_GRAY - 1; ih >= 0; ih--)
    {
      P2_sq[ih] = P2_sq[ih + 1] + data[ih + 1] * data[ih + 1];
    }

  /* Find the threshold that maximizes the criterion */
  threshold = INT_MIN;
  max_crit = DBL_MIN;
  for (it = 0; it < NUM_GRAY; it++)
    {
      crit =
	-1.0 * SAFE_LOG (P1_sq[it] * P2_sq[it]) +
	2 * SAFE_LOG (P1[it] * (1.0 - P1[it]));
      if (crit > max_crit)
	{
	  max_crit = crit;
	  threshold = it;
	}
    }

  free_histo (norm_histo);
  free (P1);
  free (P1_sq);
  free (P2_sq);

  if (img->type == PIX_GRAY8)
    return threshold;
  else
    return threshold * 256 + 128;
}


int
threshold_otsu (const Image * img)
{
  int ih;
  int threshold;
  int first_bin;		/* see below */
  int last_bin;			/* see below */
  double total_mean;		/* mean gray-level for the whole image */
  double bcv;			/* between-class variance */
  double max_bcv;		/* max BCV */
  double *cnh;			/* cumulative normalized histogram */
  double *mean;			/* mean gray-level */
  double *data;			/* normalized histogram data */
  Histogram *norm_histo;	/* normalized histogram */

  if ((img->type != PIX_GRAY8) && (img->type != PIX_GRAY16))
    {
      ERROR_RET ("Not a grayscale image !", INT_MIN);
    }

  /* Calculate the normalized histogram */
  norm_histo = normalize_histo (create_histo (img));
  if (norm_histo == NULL)
    {
      ERROR_RET ("normalize_histo() failed !", INT_MIN);
    }

  data = get_histo_data (norm_histo);

  /* Calculate the cumulative normalized histogram */
  cnh = accumulate_histo (norm_histo);

  mean = (double *) malloc (NUM_GRAY * sizeof (double));

  mean[0] = 0.0;
  for (ih = 0 + 1; ih < NUM_GRAY; ih++)
    {
      mean[ih] = mean[ih - 1] + ih * data[ih];
    }

  total_mean = mean[MAX_GRAY];

  /* 
     Determine the first non-zero 
     bin starting from the first bin 
   */
  first_bin = 0;
  for (ih = 0; ih < NUM_GRAY; ih++)
    {
      if (cnh[ih] != 0)
	{
	  first_bin = ih;
	  break;
	}
    }

  /* 
     Determine the first non-one bin 
     starting from the last bin
   */
  last_bin = MAX_GRAY;
  for (ih = MAX_GRAY; ih >= first_bin; ih--)
    {
      if (1.0 - cnh[ih] != 0)
	{
	  last_bin = ih;
	  break;
	}
    }

  /* 
     Calculate the BCV at each gray-level and
     find the threshold that maximizes it 
   */
  threshold = INT_MIN;
  max_bcv = 0.0;
  for (ih = first_bin; ih <= last_bin; ih++)
    {
      bcv = total_mean * cnh[ih] - mean[ih];
      bcv *= bcv / (cnh[ih] * (1.0 - cnh[ih]));

      if (max_bcv < bcv)
	{
	  max_bcv = bcv;
	  threshold = ih;
	}
    }

  free_histo (norm_histo);
  free (cnh);
  free (mean);

  if (img->type == PIX_GRAY8)
    return threshold;
  else
    return threshold * 256 + 128;
}

/*------------------- end of M. Emre Celebi's FOURIER 0.8  -------------------*/

void
write_pnm_file (char *filename, unsigned char *data, int depth,
		int channels, int pixels_per_line, int lines)
{
  FILE *out;
  int count;

  out = fopen (filename, "w");
  if (!out)
    {
      printf ("write_pnm_file: could nor open %s for writing: %s\n",
	      filename, strerror (errno));
      return;
    }
  if (depth == 1)
    {
      fprintf (out, "P4\n%d\n%d\n", pixels_per_line, lines);
    }
  else
    {
      fprintf (out, "P%c\n%d\n%d\n%d\n", channels == 1 ? '5' : '6',
	       pixels_per_line, lines, (int) pow (2, depth) - 1);
    }
  if (channels == 3)
    {
      for (count = 0; count < (pixels_per_line * lines * 3); count++)
	{
	  if (depth == 16)
	    fputc (*(data + 1), out);
	  fputc (*(data++), out);
	  if (depth == 16)
	    data++;
	}
    }
  else
    {
      if (depth == 1)
	{
	  pixels_per_line /= 8;
	}
      for (count = 0; count < (pixels_per_line * lines); count++)
	{
	  switch (depth)
	    {
	    case 8:
	      fputc (*(data + count), out);
	      break;
	    case 16:
	      fputc (*(data + 1), out);
	      fputc (*(data), out);
	      data += 2;
	      break;
	    default:
	      fputc (data[count], out);
	      break;
	    }
	}
    }
  fclose (out);

  return;
}


int
alloc_image (Image * alloc_img, unsigned int rows, unsigned int cols,
	     PixelType ptype)
{
  int ssize;

  alloc_img->num_rows = rows;
  alloc_img->num_cols = cols;
  alloc_img->type = ptype;
  switch (ptype)
    {
    case PIX_GRAY16:
      ssize = 2;
      break;
    case PIX_RGB8:
      ssize = 3;
      break;
    case PIX_RGB16:
      ssize = 6;
      break;
    default:
      ssize = 1;
    }
  alloc_img->data.sample16 = malloc (rows * cols * ssize);
  if (!alloc_img->data.sample16)
    {
      printf ("Cannot allocate buffer for image!\n");
      return 1;
    }
  return 0;
}


int
read_grayscale (char *filename, Image * img)
{
  FILE *infile;
  struct pam inpam;
  tuple *tuplerow;
  uint16_t *buf_ptr;
  int irow, icol;

  infile = fopen (filename, "r");
  if (infile == NULL)
    {
      printf ("%s doesnt exist?!\n", filename);
      return 1;
    }

#ifdef PAM_STRUCT_SIZE
  pnm_readpaminit (infile, &inpam, PAM_STRUCT_SIZE (tuple_type));
#else
  pnm_readpaminit (infile, &inpam, sizeof (inpam));
#endif
  if ((inpam.format != 0x5035) || (inpam.bytes_per_sample != 2))
    {
      printf ("Format not supported!\n");
      return 1;
    }
  tuplerow = pnm_allocpamrow (&inpam);


  if (alloc_image (img, inpam.height, inpam.width, PIX_GRAY16) != 0)
    return 1;

  buf_ptr = img->data.sample16;
  for (irow = 0; irow < inpam.height; irow++)
    {
      pnm_readpamrow (&inpam, tuplerow);
      for (icol = 0; icol < inpam.width; icol++)
	*buf_ptr++ = tuplerow[icol][0];
    }

  pnm_freepamrow (tuplerow);
  fclose (infile);
  return 0;
}


int
reorder_rgb (Image * red_img, Image * green_img, Image * blue_img,
	     Image * rgb_img)
{
  uint16_t *red, *green, *blue, *rgb;
  int i, itop;

  memcpy (rgb_img, red_img, sizeof (Image));
  rgb_img->type = PIX_RGB16;
  itop = rgb_img->num_rows * rgb_img->num_cols;
  rgb_img->data.sample16 = malloc (itop * 3 * sizeof (uint16_t));
  if (!rgb_img->data.sample16)
    {
      printf ("Cannot allocate buffer for RGB image!\n");
      return 1;
    }

  red = red_img->data.sample16;
  green = green_img->data.sample16;
  blue = blue_img->data.sample16;
  rgb = rgb_img->data.sample16;
  for (i = itop; i > 0; i--)
    {
      *rgb++ = *red++;
      *rgb++ = *green++;
      *rgb++ = *blue++;
    }

  return 0;
}


/** 
 * @brief Reduces red spectral overlap from an infrared image plane
 *
 * @param[in]  red_img Image pointer to 16 bit red image plane
 * @param[in]  ired_img Image pointer to 16 bit ired image plane
 * @param[out] clean_img Pointer to 16 bit grey scale image
 *
 * @return
 * - 0 if successful
 * - 1 if out of memory
 *
 * This routine is based on the observation that the relation beween the infrared value
 * ired and the red value red of an image point can be described by ired = b + a * ln (red).
 * First 1000 points are randomly chosen to calculate the linear regression coefficent a.
 * Then ired' = ired - a  * ln (red) is calculated for each pixel. Finally, the Iir' image
 * is scaled between 0 and 65535. For the logarithms a lookup table is used.
 */

int
ired_clean (Image * red_img, Image * ired_img, Image * clean_img)
{
  double rval, rsum, rrsum, risum, rfac, radd;
  double *lut_ln;
  int *calc_buf, *calc_ptr;
  int itop, irand, i;
  long int ival, isum;
  long int imin, imax;
  uint16_t *rptr, *iptr;
  uint16_t *red, *ired;

  itop = ired_img->num_rows * ired_img->num_cols;
  memcpy (clean_img, ired_img, sizeof (Image));
  clean_img->data.sample16 = malloc (itop * 2);
  lut_ln = malloc (65536 * sizeof (double));
  calc_buf = malloc (itop * sizeof (int));
  if (!clean_img->data.sample16 || !lut_ln || !calc_buf)
    {
      printf
	("ired_clean: Cannot allocate buffer memory for ired cleaning!\n");
      free (clean_img->data.sample16);
      free (lut_ln);
      free (calc_buf);
      return 1;
    }
  /* ln () lookup table */
  for (i = 0; i < 65536; i++)
    lut_ln[i] = log (i + 1);

  /* calculate linear regression ired(ln(red)) from 2000 randomly chosen points */
  red = red_img->data.sample16;
  ired = ired_img->data.sample16;
  isum = 0;
  rsum = rrsum = risum = 0;
  for (i = 2000; i > 0; i--)
    {
      irand = random () % itop;
      rval = lut_ln[red[irand]];
      ival = ired[irand];
      isum += ival;
      rsum += rval;
      rrsum += rval * rval;
      risum += rval * (double) ival;
    }
  rfac =			/* "a" in ired = b + a * ln (red) */
    (2000.0 * risum - rsum * (double) isum) / (2000.0 * rrsum - rsum * rsum);
  radd = ((double) isum - rfac * rsum) / 2000.0;	/* "b" unused */

  printf ("ired(red) = %f * ln(red) + %f\n", rfac, radd);

  /* now calculate ired' = ired - a  * ln (red) */
  imin = INT_MAX;
  imax = INT_MIN;
  rptr = red;
  iptr = ired;
  calc_ptr = calc_buf;
  for (i = itop; i > 0; i--)
    {
      rval = -lut_ln[*rptr++] * rfac + 0.5;
      ival = *iptr++ + (int) rval;
      if (ival > imax)
	imax = ival;
      if (ival < imin)
	imin = ival;
      *calc_ptr++ = ival;
    }

  /* scale the result into the output image */
  calc_ptr = calc_buf;
  iptr = clean_img->data.sample16;
  rfac = 65535.0 / (double) (imax - imin);
  for (i = itop; i > 0; i--)
    *iptr++ = (double) (*calc_ptr++ - imin) * rfac;

  free (calc_buf);
  free (lut_ln);
  return 0;
}


/** 
 * @brief 16 bit optimized mean filter
 *
 * @param[in]  in_img Pointer to 16 bit grey scale image data
 * @param[out] out_img Pointer to 16 bit grey scale image data
 * @param[in] num_rows Height of the image
 * @param[in]  num_cols Width of the image
 * @param[in]  win_rows Height of filtering window, odd
 * @param[in]  win_cols Width of filtering window, odd
 *
 * @return
 * - 0 success
 * - 1 out of memory
 *
 * @note At the image margins the size of the filtering window
 *       is adapted. So there is no need to pad the image.
 */

int
filter_mean16 (uint16_t * in_img, uint16_t * out_img, int num_rows,
	       int num_cols, int win_rows, int win_cols)
{
  uint16_t *src, *dest;
  int itop, iadd, isub;
  int ndiv, the_sum;
  int nrow, ncol;
  int hwr, hwc;
  int *sum;
  int i, j;

  if (((win_rows & 1) == 0) || ((win_cols & 1) == 0))
    {
      printf ("Mean filter window has to have odd sizes!\n");
      return 1;
    }

  sum = malloc (num_cols * sizeof (int));
  if (!sum)
    {
      printf ("Cannot allocate buffer for sums!\n");
      return 1;
    }
  dest = out_img;

  hwr = win_rows / 2;		/* half window sizes */
  hwc = win_cols / 2;

  for (j = 0; j < num_cols; j++)	/* pre-pre calculation */
    {
      sum[j] = 0;
      src = in_img + j;
      for (i = 0; i < hwr; i++)
	{
	  sum[j] += *src;
	  src += num_cols;
	}
    }

  itop = num_rows * num_cols;
  iadd = hwr * num_cols;
  isub = (hwr - win_rows) * num_cols;
  nrow = hwr;

  for (i = 0; i < num_rows; i++)
    {
      /* update row sums if possible */
      if (isub >= 0)		/* subtract old row */
	{
	  nrow--;
	  src = in_img + isub;
	  for (j = 0; j < num_cols; j++)
	    sum[j] -= *src++;
	}
      isub += num_cols;

      if (iadd < itop)		/* add new row */
	{
	  nrow++;
	  src = in_img + iadd;
	  for (j = 0; j < num_cols; j++)
	    sum[j] += *src++;
	}
      iadd += num_cols;

      /* now we do the image columns using only the precalculated sums */

      the_sum = 0;		/* precalculation */
      for (j = 0; j < hwc; j++)
	the_sum += sum[j];
      ncol = hwc;

      /* at the left margin, real index hwc lower */
      for (j = hwc; j < win_cols; j++)
	{
	  ncol++;
	  the_sum += sum[j];
	  *dest++ = the_sum / (ncol * nrow);
	}

      ndiv = ncol * nrow;
      /* in the middle, real index hwc + 1 higher */
      for (j = 0; j < num_cols - win_cols; j++)
	{
	  the_sum -= sum[j];
	  the_sum += sum[j + win_cols];
	  *dest++ = the_sum / ndiv;
	}

      /* at the right margin, real index hwc + 1 higher */
      for (j = num_cols - win_cols; j < num_cols - hwc - 1; j++)
	{
	  ncol--;
	  the_sum -= sum[j];	/* j - hwc - 1 */
	  *dest++ = the_sum / (ncol * nrow);
	}
    }
  return 0;
}

int
delta_mean (Image * in_img, uint16_t * out_img, int win_rows, int win_cols)
{
  uint16_t *src, *dest;
  int num_rows, num_cols;
  int i, itop;
  int val;

  num_rows = in_img->num_rows;
  num_cols = in_img->num_cols;
  itop = num_rows * num_cols;
  src = in_img->data.sample16;
  if (filter_mean16 (src, out_img, num_rows, num_cols,
			     win_rows, win_cols) == 0)
    {
      dest = out_img;
      for (i = 0; i < itop; i++)
	{
	  val = *dest - *src++;
	  *dest++ = abs(val);
	}
      return 0;
    }
  return 1;
}

/** 
 * @brief Find dark noise by adaptive thresholding
 *
 * @param[in]  in_img Image pointer to 16 bit grey scale image
 * @param[out] out_img Image pointer to binary image (0, 255)
 * @param[in] win_size Size of filtering window
 * @param[in]  a_val Parameter 2560 < a_val < 7680
 * @param[in]  b_val Parameter 12800 < b_val < 25600
 *
 * @return
 * - 0 success
 * - 1 out of memory
 *
 * This routine follows the concept of Crnojevic's MAD (median of the absolute deviations
 * from the median) filter. The first median filter step is replaced with a mean filter.
 * The dirty pixels which we wish to remove are always darker than the real signal. But
 * at high resolutions the scanner may generate some noise and the ired cleaning step can
 * reverse things. So a maximum filter will not do.
 * The second median is replaced by a mean filter to reduce computation time. Inspite of
 * these changes Crnojevic's recommendations for the choice of the parameters "a" and "b"
 * are still valid when scaled to 16 bit.
 *
 * @note "0" in the mask output is regarded "dirty"
 *
 * @ref Crnojevic V. (2005) "Impulse Noise Filter with Adaptive Mad-Based Threshold"
 *      Proc. of the IEEE Int. Conf. on Image Processing, 3: 337-340
 */

int
filter_madmean (Image * in_img, Image * out_img, int win_size,
	       int a_val, int b_val)
{
  uint16_t *delta_ij, *delta_ptr;
  uint16_t *mad_ij, *mad_ptr;
  uint8_t *dest8;
  double ab_term;
  int num_rows, num_cols;
  int threshold, itop;
  int ival, i;
  int result = 1;

  num_rows = in_img->num_rows;
  num_cols = in_img->num_cols;
  itop = num_rows * num_cols;

  delta_ij = malloc (itop * sizeof (uint16_t));
  mad_ij = malloc (itop * sizeof (int));
  if (delta_ij && mad_ij)
    {
      /* get the differences to the local mean */
      mad_ptr = in_img->data.sample16;
      if (filter_mean16 (mad_ptr, delta_ij, num_rows, num_cols, win_size, win_size) == 0)
	{
	  delta_ptr = delta_ij;
	  for (i = 0; i < itop; i++)
	    {
	      ival = *mad_ptr++ - *delta_ptr;
	      *delta_ptr++ = abs (ival);
	    }
	  /* make the second filtering window a bit larger */
	  win_size = ((win_size * 4) / 3) | 1;
	  /* and get the local mean differences */
	  if (filter_mean16 (delta_ij, mad_ij, num_rows, num_cols,
			     win_size, win_size) == 0)
	    {
	      mad_ptr = mad_ij;
	      delta_ptr = delta_ij;
	      dest8 = out_img->data.sample8;
	      /* construct the noise map */
	      ab_term = (b_val - a_val) / (double) b_val;
	      for (i = 0; i < itop; i++)
		{
		  /* by calculating the threshold */
		  ival = *mad_ptr++;
		  if (ival >= b_val)	/* outlier */
		    threshold = a_val;
		  else
		    threshold = a_val + (double) ival * ab_term;
		  /* above threshold is noise, indicated by 0 */
		  if (*delta_ptr++ >= threshold)
		    *dest8++ = 0;
		  else
		    *dest8++ = 255;
		}
	      result = 0;
	    }
	}
    }
  else
    printf ("filter_madmax: Cannot allocate buffers\n");

  free (mad_ij);
  free (delta_ij);
  return result;
}


int
apply_rgb_gamma (Image * rgb_img, double gamma)
{
  uint16_t *bgr, *rgb;
  double *lut_gamma;
  double di;
  int i, itop;

  lut_gamma = malloc (65536 * sizeof (double));
  if (!lut_gamma)
    {
      printf ("Cannot allocate buffer for pow() lookup!\n");
      return 1;
    }
  di = 65535 / pow (65535.0, gamma);
  for (i = 0; i < 65536; i++)
    lut_gamma[i] = di * pow ((double) i, gamma);

  itop = rgb_img->num_rows * rgb_img->num_cols * 3;
  rgb = rgb_img->data.sample16;
  bgr = rgb;
  for (i = itop; i > 0; i--)
    *rgb++ = lut_gamma[*bgr++];

  free (lut_gamma);
  return 0;
}


/** 
 * @brief Calculates minimal Manhattan distances for an image mask
 *
 * @param[in]  mask_img Image pointer to binary image (0, 255)
 * @param[out] dist_map Integer pointer to map of closest distances
 * @param[out] idx_map Integer pointer to indices of closest pixels
 * @param[in]  erode ==0: next pixel is 0, != 0: is 255
 *
 * manhattan_dist takes a mask image consisting of 0 or 255  values. Given that
 * a 0 represents a dirty pixel and erode != 0, manhattan_dist will calculate the
 * shortest distance to a clean (255) pixel and record which pixel that was so
 * that the clean parts of the image can be dilated into the dirty ones. Thresholding
 * can be done on the distance. Conversely, if erode == 0 the distance of a clean
 * pixel to the closest dirty one is calculated which can be used to dilate the mask.
 *
 * @ref extended and C version of
 *      http://ostermiller.org/dilate_and_erode.html
 */

void
manhattan_dist (Image * mask_img, unsigned int *dist_map, unsigned int *idx_map, int erode)
{
  uint8_t *mask;
  unsigned int *index, *manhattan;
  int rows, cols, itop;
  int i, j;

  if (erode != 0)
    erode = 255;

  /* initialize maps */
  rows = mask_img->num_rows;
  cols = mask_img->num_cols;
  itop = rows * cols;
  mask = mask_img->data.sample8;
  manhattan = dist_map;
  index = idx_map;
  for (i = 0; i < itop; i++)
    {
      *manhattan++ = *mask++;
      *index++ = i;
    }

  /* traverse from top left to bottom right */
  manhattan = dist_map;
  index = idx_map;
  for (i = 0; i < rows; i++)
    for (j = 0; j < cols; j++)
      {
	if (*manhattan == erode)
	  {
	    /* take original, distance = 0, index stays the same */
	    *manhattan = 0;
	  }
	else
	  {
	    /* assume maximal distance to clean pixel */
	    *manhattan = cols + rows;
	    /* or one further away than pixel to the top */
	    if (i > 0)
	      if (manhattan[-cols] + 1 < *manhattan)
		{
		  *manhattan = manhattan[-cols] + 1;
		  *index = index[-cols];	/* index follows */
		}
	    /* or one further away than pixel to the left */
	    if (j > 0)
	      {
		if (manhattan[-1] + 1 < *manhattan)
		  {
		    *manhattan = manhattan[-1] + 1;
		    *index = index[-1];	/* index follows */
		  }
		if (manhattan[-1] + 1 == *manhattan)
		  if (random () % 2 == 0)	/* chose index */
		    *index = index[-1];
	      }
	  }
	manhattan++;
	index++;
      }

  /* traverse from bottom right to top left */
  manhattan = dist_map + itop - 1;
  index = idx_map + itop - 1;
  for (i = rows - 1; i >= 0; i--)
    for (j = cols - 1; j >= 0; j--)
      {
	if (i < rows - 1)
	  {
	    /* either what we had on the first pass
	       or one more than the pixel to the bottm */
	    if (manhattan[+cols] + 1 < *manhattan)
	      {
		*manhattan = manhattan[+cols] + 1;
		*index = index[+cols];	/* index follows */
	      }
	    if (manhattan[+cols] + 1 == *manhattan)
	      if (random () % 2 == 0)	/* chose index */
		*index = index[+cols];
	  }
	if (j < cols - 1)
	  {
	    /* or one more than pixel to the right */
	    if (manhattan[1] + 1 < *manhattan)
	      {
		*manhattan = manhattan[1] + 1;
		*index = index[1];	/* index follows */
	      }
	    if (manhattan[1] + 1 == *manhattan)
	      if (random () % 2 == 0)	/* chose index */
		*index = index[1];
	  }
	manhattan--;
	index--;
      }
}


void
dilate (Image * mask_img, unsigned int *dist_map, unsigned int *idx_map, int by)
{
  uint8_t *mask;
  unsigned int *manhattan;
  int i, itop;
  int erode;

  if (by == 0)
    return;
  if (by > 0)
    erode = 0;
  else
    {
      erode = 1;
      by = -by;
    }
  itop = mask_img->num_rows * mask_img->num_cols;
  mask = mask_img->data.sample8;
  manhattan_dist (mask_img, dist_map, idx_map, erode);

  manhattan = dist_map;
  for (i = 0; i < itop; i++)
    {
      if (*manhattan++ <= by)
	*mask = 0;
      mask++;
    }

  return;
}


int
threshold_image (Image * in_img,
		 Image * mask_img, int (*thresh_func) (const Image * img),
		 int *thresh)
{
  uint16_t *in_ptr;
  uint8_t *mask_ptr;
  int threshold;
  int i;

  if (alloc_image (mask_img, in_img->num_rows, in_img->num_cols, PIX_GRAY8) !=
      0)
    return 1;

  threshold = (*thresh_func) (in_img);
  if (threshold < 0)
    {
      printf ("Thresholding failed!\n");
      return 1;
    }
  *thresh = threshold;
  printf ("Threshold = %d\n", threshold);

  in_ptr = in_img->data.sample16;
  mask_ptr = mask_img->data.sample8;
  for (i = in_img->num_rows * in_img->num_cols; i > 0; i--)
    {
      if (*in_ptr++ > threshold)
	*mask_ptr++ = 255;
      else
	*mask_ptr++ = 0;
    }
  return 0;
}


void
combine_masks (Image * comb_img, Image * an_img)
{
  uint8_t *comb_ptr, *a_ptr;
  int i, itop;

  itop = comb_img->num_rows * comb_img->num_cols;
  comb_ptr = comb_img->data.sample8;
  a_ptr = an_img->data.sample8;
  for (i = itop; i > 0; i--)
    *comb_ptr++ &= *a_ptr++;
}


/** 
 * @brief Replace dirty pixels with smoothened closest clean pixels
 *
 * @param[in]  todo Pointer to a Fill_Task structure
 *
 * @return
 * - NULL always
 *
 * This routine does the final step of dirt removal. First the clean parts of the image are
 * dilated into the dirty ones. The resulting image is run through a mean filter to adapt
 * the filled in pixels to their new surroundings. Finally the filled in pixels are replaced
 * with their smoothened counterparts.
 * If more than one processor is available the routine recursively generates more threads
 * for different color planes.
 *
 * @note The clean parts of the image remain untouched
 */

void *
dilate_soft_mt (void *todo)
{
  pthread_t child;
  Fill_Task *ours;
  Fill_Task others;
  uint16_t *color, *plane;
  unsigned int *index, *manhattan;
  int hi_plane, lo_plane;
  int rows, cols, itop;
  int dist, dist_max;
  int have_child;
  int failed;
  int rc, i;

  ours = (Fill_Task *) todo;
  rows = ours->num_rows;
  cols = ours->num_cols;
  itop = rows * cols;
  plane = malloc (itop * sizeof (uint16_t));
  if (plane)
    {
      failed = 0;

      hi_plane = ours->planes_todo - 1;
      lo_plane = 0;
      have_child = 0;
      if ((ours->cores_avail > 1) && (hi_plane > 0))
	{
	  memcpy (&others, ours, sizeof (Fill_Task));
	  others.planes_todo--;
	  others.cores_avail--;
	  others.thread_no++;
	  rc =
	    pthread_create (&child, NULL, dilate_soft_mt, (void *) &others);
	  if (rc)
	    {
	      printf
		("dilate_mean(%d): return code from pthread_create() is %d\n",
		 ours->thread_no, rc);
	      lo_plane = 0;	/* do all planes */
	    }
	  else
	    {
	      lo_plane = hi_plane;	/* do the top plane only */
	      have_child = 1;
	    }
	}

      dist_max = ours->dist_max;
      while ((hi_plane >= lo_plane) && (failed == 0))	/* in case child has failed */
	{
	  while ((hi_plane >= lo_plane) && (failed == 0))
	    {
	      /* replace dirty pixels */
	      manhattan = ours->dist_map;
	      index = ours->idx_map;
	      color = ours->in_rgb[hi_plane];
	      for (i = 0; i < itop; i++)
		{
		  dist = *manhattan++;
		  if ((dist != 0) && (dist <= dist_max))
		    color[i] = color[index[i]];
		}

	      /* generate smoothened temporary color planes */
	      failed = filter_mean16
		(color, plane, rows, cols, ours->win_size, ours->win_size);
	      if (failed)
		break;
	      else
		{
		  /* fill in smoothened pixels */
		  manhattan = ours->dist_map;
		  for (i = 0; i < itop; i++)
		    {
		      dist = *manhattan++;
		      if ((dist != 0) && (dist <= dist_max))
			color[i] = plane[i];
		    }
		}
	      hi_plane--;
	    }
	  if (have_child)
	    {
	      rc = pthread_join (child, NULL);
	      if (rc)
		{
		  printf
		    ("dilate_mean(%d): return code from pthread_join() is %d\n",
		     ours->thread_no, rc);
		  lo_plane = 0;	/* do all planes */
		}
	      else
		{
		  if (others.failed)
		    lo_plane = 0;	/* do all planes */
		}
	      have_child = 0;
	    }
	}
    }
  else
    failed = 1;

  free (plane);
  ours->failed = failed;
  return NULL;
}


int
filter_dilate_mean_mt (Image * red_img,	/* red original color plane */
		       Image * green_img,	/* same for green */
		       Image * blue_img,	/* and blue */
		       Image * mask_img,	/* contains dirty pixels */
		       Image * out_rgb_img,	/* output image RGB interleaved */
		       int dist_max,	/* maximal depth of dirt to fill */
		       int expand,	/* dilation of mask before filling */
		       int win_size)	/* window size for mean filtering */
{
  Fill_Task todo;
  uint16_t *in_rgb[3];
  unsigned int *idx_map, *dist_map;
  int rows, cols, itop;
  int failed = 0;

  rows = red_img->num_rows;
  cols = red_img->num_cols;
  itop = rows * cols;
  idx_map = malloc (itop * sizeof (unsigned int));
  dist_map = malloc (itop * sizeof (unsigned int));
  if (!idx_map || !dist_map)
    {
      printf ("Cannot allocate buffers for dilation!\n");
      failed = 1;
    }
  else
    {
      if (expand > 0)
	dilate (mask_img, dist_map, idx_map, expand);
      manhattan_dist (mask_img, dist_map, idx_map, 1);

      /* replace dirty pixels */
      in_rgb[0] = red_img->data.sample16;
      in_rgb[1] = green_img->data.sample16;
      in_rgb[2] = blue_img->data.sample16;
      todo.in_rgb = in_rgb;
      todo.idx_map = idx_map;
      todo.dist_map = dist_map;
      todo.num_rows = rows;
      todo.num_cols = cols;
      todo.win_size = win_size;
      todo.dist_max = dist_max;
      todo.planes_todo = 3;
      todo.thread_no = 0;
      todo.cores_avail = get_nprocs ();
      dilate_soft_mt ((void *) &todo);

      /* finally compose to RGB image */
      if ((todo.failed == 0)
	  && (reorder_rgb (red_img, green_img, blue_img, out_rgb_img) == 0))
	failed = 0;
      else
	failed = 1;
    }

  free (dist_map);
  free (idx_map);
  return failed;
}


int
filter_dilate_mean (Image * red_img,	/* red original color plane */
		    Image * green_img,	/* same for green */
		    Image * blue_img,	/* and blue */
		    Image * mask_img,	/* contains dirty pixels */
		    Image * out_rgb_img,	/* output image RGB interleaved */
		    int dist_max,	/* maximal depth of dirt to fill */
		    int expand,	/* dilation of mask before filling */
		    int win_size)	/* window size for mean filtering */
{
  uint16_t *in_rgb[3], *rgb_map[3], *out_rgb;
  unsigned int *idx_map, *dist_map;
  unsigned int *index, *manhattan;
  int rows, cols;
  int itop, dist;
  int i, j, k;

  rows = red_img->num_rows;
  cols = red_img->num_cols;
  itop = rows * cols;
  if (alloc_image (out_rgb_img, rows, cols, PIX_RGB16) != 0)
    return 1;

  rgb_map[0] = malloc (itop * sizeof (uint16_t));
  rgb_map[1] = malloc (itop * sizeof (uint16_t));
  rgb_map[2] = malloc (itop * sizeof (uint16_t));
  idx_map = malloc (itop * sizeof (unsigned int));
  dist_map = malloc (itop * sizeof (unsigned int));
  if (!rgb_map[0] || !rgb_map[1] || !rgb_map[2] || !idx_map || !dist_map)
    {
      printf ("Cannot allocate buffers for dilation!\n");
      free (out_rgb_img->data.sample16);
      free (rgb_map[0]);
      free (rgb_map[1]);
      free (rgb_map[2]);
      free (idx_map);
      free (dist_map);
      return 1;
    }

  if (expand > 0)
    dilate (mask_img, dist_map, idx_map, expand);
  manhattan_dist (mask_img, dist_map, idx_map, 1);

  /* replace dirty pixels */
  in_rgb[0] = red_img->data.sample16;
  in_rgb[1] = green_img->data.sample16;
  in_rgb[2] = blue_img->data.sample16;

  manhattan = dist_map;
  index = idx_map;
  for (i = 0; i < itop; i++)
    {
      dist = *manhattan++;
      if ((dist != 0) && (dist <= dist_max))
	{
	  j = index[i];
	  for (k = 0; k < 3; k++)
	    in_rgb[k][i] = in_rgb[k][j];
	}
    }

  /* generate smoothened temporary color planes */
  for (k = 0; k < 3; k++)
    {
      if (filter_mean16
	  (in_rgb[k], rgb_map[k], rows, cols, win_size, win_size) != 0)
	{
	  free (rgb_map[0]);
	  free (rgb_map[1]);
	  free (rgb_map[2]);
	  free (dist_map);
	  free (idx_map);
	  return 1;
	}
    }

  /* generate final RGB image taking original clean pixels 
     or replaced and smoothened ones */
  manhattan = dist_map;
  out_rgb = out_rgb_img->data.sample16;
  for (i = 0; i < itop; i++)
    {
      dist = *manhattan++;
      if ((dist != 0) && (dist <= dist_max))
	for (k = 0; k < 3; k++)
	  *out_rgb++ = rgb_map[k][i];
      else
	for (k = 0; k < 3; k++)
	  *out_rgb++ = in_rgb[k][i];
    }

  free (rgb_map[0]);
  free (rgb_map[1]);
  free (rgb_map[2]);
  free (dist_map);
  free (idx_map);
  return 0;
}


int
main (int argc, char **argv)
{
  struct timeval start, end;
  Image red_img, green_img, blue_img, ired_img, rgb_img;
  Image clean_img, thresh_yen_img, thresh_adapt_img;
  Image dilate_img;
  int thresh_yen, thresh_otsu;
  int resolution, winsize_smooth;
  int winsize_filter, size_dilate;

  if (argc > 1)
    resolution = atoi (argv[1]);
  else
    {
      printf ("Usage: ircleanest <resolution>, e.g. ircleanest 1200\n");
      return 1;
    }

  gettimeofday(&start, NULL);

  /* setup resolution dependent parameters */
  winsize_smooth = (resolution / 540) | 1;
  if (winsize_smooth < 3)
    winsize_smooth = 3;
/*  winsize_filter = (int) (5.0 * sqrt (resolution / 300.0)) | 1;*/
  winsize_filter = (int) (5.0 * (double) resolution / 300.0) | 1;
  if (winsize_filter < 3)
    winsize_filter = 3;
  size_dilate = resolution / 1500;

  /* reduce spectral red component in ired plane */
  if (read_grayscale (RED_NAME, &red_img) != 0)
    return 1;
  if (read_grayscale (IRED_NAME, &ired_img) != 0)
    return 1;
  if ((red_img.num_cols !=
       ired_img.num_cols) || (red_img.num_rows != ired_img.num_rows))
    {
      printf ("Images of unequal dimensions!\n");
      return 1;
    }
  if (ired_clean (&red_img, &ired_img, &clean_img) != 0)
    return 1;
  write_pnm_file (CLEAN_NAME,
		  clean_img.data.sample8, 16, 1,
		  clean_img.num_cols, clean_img.num_rows);

  /* generate a "bimodal" static threshold */
  if (threshold_image
      (&clean_img, &thresh_yen_img, threshold_yen, &thresh_yen) != 0)
    return 1;
  write_pnm_file (THRESH_1_NAME,
		  thresh_yen_img.data.sample8, 8, 1,
		  thresh_yen_img.num_cols, thresh_yen_img.num_rows);
  /* generate Otsu static threshold */
  if (threshold_image
      (&clean_img, &thresh_adapt_img, threshold_otsu, &thresh_otsu) != 0)
    return 1;
  write_pnm_file (THRESH_2_NAME,
		  thresh_adapt_img.data.sample8, 8, 1,
		  thresh_adapt_img.num_cols, thresh_adapt_img.num_rows);

  /* generate dirt mask with our own adaptive thresholding filter and add the
     dirt from the Yen's static thresholding mask */
  if (alloc_image (&thresh_adapt_img, red_img.num_rows, red_img.num_cols, PIX_GRAY8) !=
      0)
    return 1;
  if (filter_madmean (&clean_img, &thresh_adapt_img, winsize_filter, 5120, 25600) /* 5120, 25600*/
      != 0)
    /* last two parameters: 2560, 9600 detects more, 5120, 19200 less */
    return 1;
  write_pnm_file (MADM_NAME,
		  thresh_adapt_img.data.sample8, 8, 1,
		  thresh_adapt_img.num_cols, thresh_adapt_img.num_rows);
  combine_masks (&thresh_adapt_img, &thresh_yen_img);	/* add large black dirt */
  free (ired_img.data.sample16);	/* ired plane */
  free (clean_img.data.sample16);	/* cleaned ired */
  free (thresh_yen_img.data.sample16);	/* mask */


  /* generate the original RGB image */
  if (read_grayscale (GREEN_NAME, &green_img) != 0)
    return 1;
  if (read_grayscale (BLUE_NAME, &blue_img) != 0)
    return 1;

  /* only for display */
  if (reorder_rgb (&red_img, &green_img, &blue_img, &rgb_img) != 0)
    return 1;
  write_pnm_file (RGB1_NAME,
		  rgb_img.data.sample8, 16, 3,
		  rgb_img.num_cols, rgb_img.num_rows);
  free (rgb_img.data.sample16);


  if (filter_dilate_mean_mt
      (&red_img, &green_img, &blue_img, &thresh_adapt_img, &dilate_img, 500,
       size_dilate, winsize_smooth) != 0)
    return 1;
  write_pnm_file (RGB3_NAME,
		  dilate_img.data.sample8, 16, 3,
		  dilate_img.num_cols, dilate_img.num_rows);


  free (thresh_adapt_img.data.sample16);


  if (apply_rgb_gamma (&dilate_img, 0.454545455) != 0)
    return 1;
  write_pnm_file (RGB4_NAME,
		  dilate_img.data.sample8, 16, 3,
		  dilate_img.num_cols, dilate_img.num_rows);

  free (green_img.data.sample16);
  free (dilate_img.data.sample16);

  gettimeofday(&end, NULL);
  printf("Elapsed %ld ms\n", ((end.tv_sec * 1000000 + end.tv_usec)
		  - (start.tv_sec * 1000000 + start.tv_usec))/1000);

  return 0;
}

